/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/
/*
	Cube texture loading class

*/


#pragma once

#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


#include "helpers.h"


extern VkCoffeeDevice *				coffeeDevice;
extern VkCoffeeSwapChain*			coffeeSwapChain;
extern VkCommandPool				cmdPool;

class VkCoffeeTGATextureCube
{
public:
	VkSampler				sampler			    = NULL;
	VkImage					image			    = NULL;
	VkImageView				imageView		    = NULL;
	VkDeviceMemory			imageMemory		    = { 0 };
	VkDescriptorImageInfo	descriptorImageInfo;
	int						width			    = 0;
	int						height			    = 0;
	const int				numCubeSides		= 6;

	struct TGAFILE
	{
		unsigned char  imageTypeCode;
		short int      imageWidth;
		short int      imageHeight;
		unsigned char  bitCount;
		unsigned char *imageData;
	};


	VkCoffeeTGATextureCube(	VkDevice			device,
							VkPhysicalDevice	physicalDevice)
	{
		// Temp - hard code width/height of the cube map images
		width  = 512;
		height = 512;

		VkSamplerCreateInfo samplerCreateInfo	= {};
		samplerCreateInfo.sType					= VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
		samplerCreateInfo.magFilter				= VK_FILTER_LINEAR;
		samplerCreateInfo.minFilter				= VK_FILTER_LINEAR;
		samplerCreateInfo.mipmapMode			= VK_SAMPLER_MIPMAP_MODE_LINEAR;
		samplerCreateInfo.addressModeU			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeV			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeW			= VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.mipLodBias			= 0;
		samplerCreateInfo.anisotropyEnable		= VK_FALSE;
		samplerCreateInfo.minLod				= 0;
		samplerCreateInfo.maxLod				= 5;
		samplerCreateInfo.borderColor			= VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
		samplerCreateInfo.unnormalizedCoordinates = VK_FALSE;
		samplerCreateInfo.anisotropyEnable		= VK_TRUE;

    
		VkResult result = 
		vkCreateSampler( device, 
						 &samplerCreateInfo, 
						 NULL, 
						 &sampler );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create sampler." );





		VkMemoryAllocateInfo memAllocInfo {};
		memAllocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		VkMemoryRequirements memReqs;

#if 1
		// Create a host-visible staging buffer that contains the raw image data
		VkBuffer stagingBuffer;
		VkDeviceMemory stagingMemory;
#endif

		VkBufferCreateInfo bufferCreateInfo {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.size = 512 * 512 * 4 * numCubeSides;
		// This buffer is used as a transfer source for the buffer copy
		bufferCreateInfo.usage = VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
		bufferCreateInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

#if 1
		VK_CHECK_RESULT(vkCreateBuffer(device, &bufferCreateInfo, nullptr, &stagingBuffer));

		// Get memory requirements for the staging buffer (alignment, memory type bits)
		vkGetBufferMemoryRequirements(device, stagingBuffer, &memReqs);
		memAllocInfo.allocationSize = memReqs.size;
#endif



		unsigned char* imageData   = (unsigned char *) new DWORD[ width * height * numCubeSides ];

		const char* tgafiles[] = {
		"assets/rockyvalley_rt.tga",
		"assets/rockyvalley_lf.tga",		
		"assets/rockyvalley_up.tga",
		"assets/rockyvalley_dn.tga",
		"assets/rockyvalley_bk.tga",
		"assets/rockyvalley_ft.tga",
		};

		int offs = 0;
		for (int i=0; i<numCubeSides; ++i)
		{
			TGAFILE tgaFile;
			bool err =
			LoadTGAFile( tgafiles[i], &tgaFile );
			DBG_ASSERT(err);


			DBG_ASSERT(tgaFile.imageWidth == 512 );
			DBG_ASSERT(tgaFile.imageHeight == 512 );

			DBG_ASSERT(tgaFile.bitCount == 32 );
			unsigned char* imageDataTGA = (unsigned char*)tgaFile.imageData;

			memcpy( imageData+offs, imageDataTGA, 512*512*4 );
			offs += 512*512*4;
		}

#if 1
		memAllocInfo.memoryTypeIndex = GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT);

		VK_CHECK_RESULT(vkAllocateMemory(device, &memAllocInfo, nullptr, &stagingMemory));
		VK_CHECK_RESULT(vkBindBufferMemory(device, stagingBuffer, stagingMemory, 0));

		// Copy texture data into staging buffer
		uint8_t *data;
		VK_CHECK_RESULT(vkMapMemory(device, stagingMemory, 0, memReqs.size, 0, (void **)&data));
		memcpy(data, imageData, 512*512*4 * numCubeSides);
		vkUnmapMemory(device, stagingMemory);

		// Create optimal tiled target image
		VkImageCreateInfo imageCreateInfo {};
		imageCreateInfo.sType			= VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		imageCreateInfo.imageType		= VK_IMAGE_TYPE_2D;
		imageCreateInfo.format			= VK_FORMAT_R8G8B8A8_UNORM;
		imageCreateInfo.mipLevels		= 1;
		imageCreateInfo.samples			= VK_SAMPLE_COUNT_1_BIT;
		imageCreateInfo.tiling			= VK_IMAGE_TILING_OPTIMAL;
		imageCreateInfo.usage			= VK_IMAGE_USAGE_SAMPLED_BIT;
		imageCreateInfo.sharingMode		= VK_SHARING_MODE_EXCLUSIVE;
		imageCreateInfo.initialLayout	= VK_IMAGE_LAYOUT_UNDEFINED;
		imageCreateInfo.extent			= {512, 512, 1};
		imageCreateInfo.usage			= VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;
		// Cube faces count as array layers in Vulkan
		imageCreateInfo.arrayLayers		= numCubeSides;
		// This flag is required for cube map images
		imageCreateInfo.flags			= VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT;

		VK_CHECK_RESULT(vkCreateImage(device, &imageCreateInfo, nullptr, &image));

		vkGetImageMemoryRequirements(device, image, &memReqs);

		memAllocInfo.allocationSize = memReqs.size;
		memAllocInfo.memoryTypeIndex = GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

		VK_CHECK_RESULT(vkAllocateMemory(device, &memAllocInfo, nullptr, &imageMemory));
		VK_CHECK_RESULT(vkBindImageMemory(device, image, imageMemory, 0));

		//==============================

		VkCommandBuffer copyCmd = CreateCommandBuffer();

		VkCommandBufferBeginInfo cmdBufferBeginInfo {};
		cmdBufferBeginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;

		VK_CHECK_RESULT(vkBeginCommandBuffer(copyCmd, &cmdBufferBeginInfo));


		// Setup buffer copy regions for each face including all of it's miplevels
		std::vector<VkBufferImageCopy> bufferCopyRegions;
		uint32_t offset = 0;

		for (uint32_t face = 0; face < (uint32_t)numCubeSides; ++face)
		{
			VkBufferImageCopy bufferCopyRegion = {};
			bufferCopyRegion.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
			bufferCopyRegion.imageSubresource.mipLevel = 0;
			bufferCopyRegion.imageSubresource.baseArrayLayer = face;
			bufferCopyRegion.imageSubresource.layerCount = 1;
			bufferCopyRegion.imageExtent.width =  512;
			bufferCopyRegion.imageExtent.height = 512;
			bufferCopyRegion.imageExtent.depth = 1;
			bufferCopyRegion.bufferOffset = offset;

			bufferCopyRegions.push_back(bufferCopyRegion);

			// Increase offset into staging buffer for next face
			offset += 512*512*4;
		}

		// Image barrier for optimal image (target)
		// Set initial layout for all array layers (faces) of the optimal (target) tiled texture
		VkImageSubresourceRange subresourceRange = {};
		subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
		subresourceRange.baseMipLevel = 0;
		subresourceRange.levelCount = 1;
		subresourceRange.layerCount = numCubeSides;

		
		SetImageLayout(
			copyCmd,
			image,
			VK_IMAGE_ASPECT_COLOR_BIT,
			VK_IMAGE_LAYOUT_UNDEFINED,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			subresourceRange);

		// Copy the cube map faces from the staging buffer to the optimal tiled image
		vkCmdCopyBufferToImage(
			copyCmd,
			stagingBuffer,
			image,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			static_cast<uint32_t>(bufferCopyRegions.size()),
			bufferCopyRegions.data()
			);

		// Change texture image layout to shader read after all faces have been copied
		VkImageLayout imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		SetImageLayout(
			copyCmd,
			image,
			VK_IMAGE_ASPECT_COLOR_BIT,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			imageLayout,
			subresourceRange);

		// Flush
		{
			VK_CHECK_RESULT(vkEndCommandBuffer(copyCmd));

			VkSubmitInfo submitInfo = {};
			submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
			submitInfo.commandBufferCount = 1;
			submitInfo.pCommandBuffers = &copyCmd;

			
			VK_CHECK_RESULT(vkQueueSubmit(coffeeDevice->queue, 1, &submitInfo, VK_NULL_HANDLE));
			VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));

			vkFreeCommandBuffers(device, cmdPool, 1, &copyCmd);
		}
#endif

		// Create image view
		VkImageViewCreateInfo view {};
		view.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		// Cube map view type
		view.viewType = VK_IMAGE_VIEW_TYPE_CUBE;
		view.format = VK_FORMAT_R8G8B8A8_UNORM;
		view.components = { VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A };
		view.subresourceRange = { VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 };
		// 6 array layers (faces)
		view.subresourceRange.layerCount = numCubeSides;
		// Set number of mip levels
		view.subresourceRange.levelCount = 1;//cubeMap.mipLevels;
		view.image = image;
		VK_CHECK_RESULT(vkCreateImageView(device, &view, nullptr,  &imageView));

#if 1
		// Clean up staging resources
		vkFreeMemory(device, stagingMemory, nullptr);
		vkDestroyBuffer(device, stagingBuffer, nullptr);
#endif

		// Fill descriptor image info that can be used for setting up descriptor sets
		descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
		descriptorImageInfo.imageView   = imageView;
		descriptorImageInfo.sampler     = sampler;
	}



	uint32_t GetMemoryType(uint32_t typeBits, VkMemoryPropertyFlags properties, VkBool32 *memTypeFound = nullptr)
	{
		// read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( coffeeDevice->physicalDevice, &memoryProperties );

		for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
		{
			if ((typeBits & 1) == 1)
			{
				if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
				{
					if (memTypeFound)
					{
						*memTypeFound = true;
					}
					return i;
				}
			}
			typeBits >>= 1;
		}
		DBG_ASSERT(false);
		return 0;
	}


	void SetImageLayout(
		VkCommandBuffer cmdbuffer, 
		VkImage image, 
		VkImageAspectFlags aspectMask, 
		VkImageLayout oldImageLayout, 
		VkImageLayout newImageLayout,
		VkImageSubresourceRange subresourceRange,
		VkPipelineStageFlags srcStageMask = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
		VkPipelineStageFlags dstStageMask = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT)
	{
		// Create an image barrier object
		VkImageMemoryBarrier imageMemoryBarrier {};
		imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
		imageMemoryBarrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
		imageMemoryBarrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
		imageMemoryBarrier.oldLayout = oldImageLayout;
		imageMemoryBarrier.newLayout = newImageLayout;
		imageMemoryBarrier.image = image;
		imageMemoryBarrier.subresourceRange = subresourceRange;

		// Source layouts (old)
		// Source access mask controls actions that have to be finished on the old layout
		// before it will be transitioned to the new layout
		switch (oldImageLayout)
		{
		case VK_IMAGE_LAYOUT_UNDEFINED:
				// Image layout is undefined (or does not matter)
				// Only valid as initial layout
				// No flags required, listed only for completeness
				imageMemoryBarrier.srcAccessMask = 0;
				break;

		case VK_IMAGE_LAYOUT_PREINITIALIZED:
				// Image is preinitialized
				// Only valid as initial layout for linear images, preserves memory contents
				// Make sure host writes have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_HOST_WRITE_BIT;
				break;

		case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL:
				// Image is a color attachment
				// Make sure any writes to the color buffer have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
				break;

		case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL:
				// Image is a depth/stencil attachment
				// Make sure any writes to the depth/stencil buffer have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
				break;

		case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL:
				// Image is a transfer source 
				// Make sure any reads from the image have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
				break;

		case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
				// Image is a transfer destination
				// Make sure any writes to the image have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
				break;

		case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
				// Image is read by a shader
				// Make sure any shader reads from the image have been finished
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_SHADER_READ_BIT;
				break;
		}

		// Target layouts (new)
		// Destination access mask controls the dependency for the new image layout
		switch (newImageLayout)
		{
		case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
			// Image will be used as a transfer destination
			// Make sure any writes to the image have been finished
			imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
			break;

		case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL:
			// Image will be used as a transfer source
			// Make sure any reads from and writes to the image have been finished
			imageMemoryBarrier.srcAccessMask = imageMemoryBarrier.srcAccessMask | VK_ACCESS_TRANSFER_READ_BIT;
			imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
			break;

		case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL:
			// Image will be used as a color attachment
			// Make sure any writes to the color buffer have been finished
			imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
			imageMemoryBarrier.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
			break;

		case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL:
			// Image layout will be used as a depth/stencil attachment
			// Make sure any writes to depth/stencil buffer have been finished
			imageMemoryBarrier.dstAccessMask = imageMemoryBarrier.dstAccessMask | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
			break;

		case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
			// Image will be read in a shader (sampler, input attachment)
			// Make sure any writes to the image have been finished
			if (imageMemoryBarrier.srcAccessMask == 0)
			{
				imageMemoryBarrier.srcAccessMask = VK_ACCESS_HOST_WRITE_BIT | VK_ACCESS_TRANSFER_WRITE_BIT;
			}
			imageMemoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
			break;
		}

		// Put barrier inside setup command buffer
		vkCmdPipelineBarrier(
			cmdbuffer, 
			srcStageMask, 
			dstStageMask, 
			0, 
			0, nullptr,
			0, nullptr,
			1, &imageMemoryBarrier);
	}

	VkCommandBuffer CreateCommandBuffer()
	{
		

		// Create one command buffer for each swap chain image and reuse for rendering
		VkCommandBuffer cmdBuffer;

		VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
		cmdBufAllocateInfo.sType              = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool        = cmdPool;
		cmdBufAllocateInfo.level              = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount = 1;//static_cast<uint32_t>(clearCmdBuffers.size());

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, &cmdBuffer));

		return cmdBuffer;
	}


	bool LoadTGAFile(const char *filename, TGAFILE *tgaFile)
	{
		FILE *filePtr;
		unsigned char ucharBad;
		short int sintBad;
		long imageSize;
		int colorMode;
		unsigned char colorSwap;

		// Open the TGA file.
		filePtr = fopen(filename, "rb");
		if (filePtr == NULL)
		{
			return false;
		}

		// Read the two first bytes we don't need.
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);

		// Which type of image gets stored in imageTypeCode.
		fread(&tgaFile->imageTypeCode, sizeof(unsigned char), 1, filePtr);

		// For our purposes, the type code should be 2 (uncompressed RGB image)
		// or 3 (uncompressed black-and-white images).
		if (tgaFile->imageTypeCode != 2 && tgaFile->imageTypeCode != 3)
		{
			fclose(filePtr);
			return false;
		}

		// Read 13 bytes of data we don't need.
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);

		// Read the image's width and height.
		fread(&tgaFile->imageWidth, sizeof(short int), 1, filePtr);
		fread(&tgaFile->imageHeight, sizeof(short int), 1, filePtr);

		// Read the bit depth.
		fread(&tgaFile->bitCount, sizeof(unsigned char), 1, filePtr);

		// Read one byte of data we don't need.
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);

		// Color mode -> 3 = BGR, 4 = BGRA.
		colorMode = tgaFile->bitCount / 8;
		imageSize = tgaFile->imageWidth * tgaFile->imageHeight * colorMode;

		// Allocate memory for the image data.
		tgaFile->imageData = (unsigned char*)malloc(sizeof(unsigned char)*imageSize);

		// Read the image data.
		fread(tgaFile->imageData, sizeof(unsigned char), imageSize, filePtr);

		// Change from BGR to RGB so OpenGL can read the image data.
		for (int imageIdx = 0; imageIdx < imageSize; imageIdx += colorMode)
		{
			colorSwap = tgaFile->imageData[imageIdx];
			tgaFile->imageData[imageIdx] = tgaFile->imageData[imageIdx + 2];
			tgaFile->imageData[imageIdx + 2] = colorSwap;
		}

		fclose(filePtr);
		return true;
	}

};